# A NixOS module which can run a Ristretto-based issuer for PrivateStorage
# ZKAPs.
{ lib, pkgs, ourpkgs, config, ... }: let
  cfg = config.services.private-storage-issuer;
in {
  options = {
    services.private-storage-issuer.enable = lib.mkEnableOption "PrivateStorage ZKAP Issuer Service";
    services.private-storage-issuer.package = lib.mkOption {
      default = ourpkgs.zkapissuer;
      type = lib.types.package;
      example = lib.literalExpression "pkgs.zkapissuer.components.exes.\"PaymentServer-exe\"";
      description = ''
        The package to use for the ZKAP issuer.
      '';
    };
    services.private-storage-issuer.domains = lib.mkOption {
      type = lib.types.listOf lib.types.str;
      example = [ "payments.example.com" ];
      description = ''
        The domain names at which the issuer is reachable.
      '';
    };
    services.private-storage-issuer.tls = lib.mkOption {
      default = true;
      type = lib.types.bool;
      description = ''
        Whether or not to listen on TLS.  For real-world use you should always
        listen on TLS.  This is provided as an aid to automated testing where
        it might be difficult to obtain a real certificate.
      '';
    };
    services.private-storage-issuer.issuer = lib.mkOption {
      default = "Ristretto";
      type = lib.types.enum [ "Trivial" "Ristretto" ];
      example = "Trivial";
      description = ''
        The issuer algorithm to use.  Either Trivial for a fake no-crypto
        algorithm or Ristretto for Ristretto-flavored PrivacyPass.
      '';
    };
    services.private-storage-issuer.tokensPerVoucher = lib.mkOption {
      default = null;
      type = lib.types.nullOr lib.types.int;
      example = 50000;
      description = ''
        If not null, a value to pass to PaymentServer for
        ``--tokens-per-voucher``.
      '';
    };
    services.private-storage-issuer.ristrettoSigningKeyPath = lib.mkOption {
      default = null;
      type = lib.types.path;
      description = ''
        The path to a file containing the Ristretto signing key to use.
        Required if the issuer is ``Ristretto``.
      '';
    };
    services.private-storage-issuer.stripeSecretKeyPath = lib.mkOption {
      type = lib.types.path;
      description = ''
        The path to a file containing a Stripe secret key to use for charge
        and payment management.
      '';
    };
    services.private-storage-issuer.stripeWebhookSecretKeyPath = lib.mkOption {
      type = lib.types.path;
      description = ''
        The path to a file containing a Stripe "webhook" secret key to use for
        charge and payment management.
      '';
    };
    services.private-storage-issuer.stripeEndpointDomain = lib.mkOption {
      type = lib.types.str;
      description = ''
        The domain name for the Stripe API HTTP endpoint.
      '';
      default = "api.stripe.com";
    };
    services.private-storage-issuer.stripeEndpointScheme = lib.mkOption {
      type = lib.types.enum [ "HTTP" "HTTPS" ];
      description = ''
        Whether to use HTTP or HTTPS for the Stripe API.
      '';
      default = "HTTPS";
    };
    services.private-storage-issuer.stripeEndpointPort = lib.mkOption {
      type = lib.types.int;
      description = ''
        The port number for the Stripe API HTTP endpoint.
      '';
      default = 443;
    };
    services.private-storage-issuer.database = lib.mkOption {
      default = "Memory";
      type = lib.types.enum [ "Memory" "SQLite3" ];
      description = ''
        The kind of voucher database to use.
      '';
    };
    services.private-storage-issuer.databaseFileSystem = lib.mkOption {
      # Logically, the type is the type of an entry in fileSystems - but we'll
      # just let the type system enforce that when we pass the value on to
      # fileSystems.
      description = ''
        Configuration for a filesystem to mount which will hold the issuer's
        internal state database.
      '';
    };
    services.private-storage-issuer.databasePath = lib.mkOption {
      default = null;
      type = lib.types.str;
      description = ''
        The path to a database file in the filesystem, if the SQLite3 database
        type is being used.
      '';
    };
    services.private-storage-issuer.letsEncryptAdminEmail = lib.mkOption {
      type = lib.types.str;
      description = ''
        An email address to give to Let's Encrypt as an operational contact
        for the service's TLS certificate.
      '';
    };
    services.private-storage-issuer.allowedChargeOrigins = lib.mkOption {
      type = lib.types.listOf lib.types.str;
      description = ''
        The CORS "Origin" values which are allowed to submit charges to the
        payment server.  Note this is not currently enforced by the
        PaymentServer.  It just controls the CORS headers served.
      '';
    };
  };

  config =
    let
      # We'll refer to this collection of domains by the first domain in the
      # list.
      domain = builtins.head cfg.domains;
      # Payment server internal http port (arbitrary, non-priviledged):
      internalHttpPort = "1061";

      # The "-vN" suffix indicates that this Nth incompatible version of on
      # disk state as managed by this deployment system.  This does not have
      # anything to do with what's inside the PaymentServer-managed state.
      # Instead it's about things like the type of filesystem used or options
      # having to do with the backing volume behind the filesystem.  In
      # general I expect that to get from "-vN" to "-v(N+1)" some manual
      # upgrade steps will be required.
      stateDirectory = "zkapissuer-v2";

    in lib.mkIf cfg.enable {
    # Make sure the voucher database filesystem is mounted.
    fileSystems = {
      "zkapissuer-data" = cfg.databaseFileSystem // {
        mountPoint = "/var/lib/${stateDirectory}";
      };
    };

    # Add a systemd service to run PaymentServer.
    systemd.services.zkapissuer = {
      enable = true;
      description = "ZKAP Issuer";
      wantedBy = [ "multi-user.target" ];

      # It really shouldn't ever exit on its own!  If it does, it's a bug
      # we'll have to fix.  Restart it and hope it doesn't happen too much
      # before we can fix whatever the issue is.
      serviceConfig.Restart = "always";
      serviceConfig.Type = "simple";

      # Run w/o privileges
      serviceConfig = {
        DynamicUser = false;
        User = "zkapissuer";
        Group = "zkapissuer";
      };

      # Make systemd create a User/Group owned directory for PaymentServer
      # state. According to the docs at
      # https://www.freedesktop.org/software/systemd/man/systemd.exec.html#RuntimeDirectory=
      # "The specified directory names must be relative" ... this makes
      # systemd create this directory in /var/lib/ for us.
      serviceConfig.StateDirectory = stateDirectory;
      serviceConfig.StateDirectoryMode = "0750";

      unitConfig.AssertPathExists = [
        # Bail if there is still an old (root-owned) DB file on this system.
        # If you hit this, and this /var/db/ file is indeed current, move it
        # to /var/lib/zkapissuer/vouchers.sqlite3 and chown it to
        # zkapissuer:zkapissuer.
        "!/var/db/vouchers.sqlite3"

        # Similarly, bail if the newer path you were just told to create --
        # /var/lib/zkapissuer/vouchers.sqlite3 -- exists.  It needs to be
        # moved /var/lib/zkapissuer-v2 where a dedicated filesystem has been
        # created for it.
        "!/var/lib/zkapissuer/vouchers.sqlite3"
      ];

      # Only start if the dedicated vouchers database filesystem is mounted so
      # that we know we're going to find our vouchers database there (or that
      # we will create it in the right place).
      unitConfig.Requires = ["local-fs.target"];
      unitConfig.After = ["local-fs.target"];

      script =
        let
          # Compute the right command line arguments to pass to it.  The
          # signing key is only supplied when using the Ristretto issuer.
          issuerArgs =
            if cfg.issuer == "Trivial"
              then "--issuer Trivial"
              else "--issuer Ristretto --signing-key-path ${cfg.ristrettoSigningKeyPath}";
          databaseArgs =
            if cfg.database == "Memory"
              then "--database Memory"
              else "--database SQLite3 --database-path ${cfg.databasePath}";
          httpArgs = "--http-port ${internalHttpPort}";

          prefixOption = s: "--cors-origin=" + s;
          originStrings = map prefixOption cfg.allowedChargeOrigins;
          originArgs = builtins.concatStringsSep " " originStrings;

          stripeArgs =
            "--stripe-key-path ${cfg.stripeSecretKeyPath} " +
            "--stripe-webhook-key-path ${cfg.stripeWebhookSecretKeyPath} " +
            "--stripe-endpoint-domain ${cfg.stripeEndpointDomain} " +
            "--stripe-endpoint-scheme ${cfg.stripeEndpointScheme} " +
            "--stripe-endpoint-port ${toString cfg.stripeEndpointPort}";

          redemptionConfig = lib.optionalString (cfg.tokensPerVoucher != null) "--tokens-per-voucher ${builtins.toString cfg.tokensPerVoucher}";
        in
          "${cfg.package.exePath} ${originArgs} ${issuerArgs} ${databaseArgs} ${httpArgs} ${stripeArgs} ${redemptionConfig}";
    };

    # PaymentServer runs as this user and group by default
    # Mind the comments in nixpkgs/nixos/modules/misc/ids.nix: "When adding a uid,
    # make sure it doesn't match an existing gid. And don't use uids above 399!"
    ids.uids.zkapissuer = 397;
    ids.gids.zkapissuer = 397;
    users.extraGroups.zkapissuer.gid = config.ids.gids.zkapissuer;
    users.extraUsers.zkapissuer = {
      uid = config.ids.uids.zkapissuer;
      isNormalUser = false;
      group = "zkapissuer";
      # Let PaymentServer read from keys, if necessary.
      extraGroups = [ "keys" ];
    };

    # Open 80 and 443 for nginx
    networking.firewall.allowedTCPPorts = [
      80
      443
    ];

    # NGINX reverse proxy
    security.acme.email = cfg.letsEncryptAdminEmail;
    security.acme.acceptTerms = true;
    services.nginx = {
      enable = true;

      recommendedGzipSettings = true;
      recommendedOptimisation = true;
      recommendedProxySettings = true;
      recommendedTlsSettings = true;

      virtualHosts."${domain}" = {
        serverAliases = builtins.tail cfg.domains;
        enableACME = cfg.tls;
        forceSSL = cfg.tls;
        locations."/v1/" = {
          # Only forward requests beginning with /v1/ so
          # we pass less scanning spam on to our backend
          # Want a regex instead? try locations."~ /v\d+/"
          proxyPass = "http://127.0.0.1:${internalHttpPort}";

          # The redemption endpoint can intentionally delay its response for
          # up to 600 seconds for a cheap kind of server-push when payment
          # completes.  Let that timeout control how long the connection stays
          # open.  PaymentServer does not accept configuration for that
          # duration so we also hard-code it here.
          #
          # http://nginx.org/en/docs/http/ngx_http_proxy_module.html#proxy_read_timeout
          extraConfig = ''
            proxy_read_timeout 660;
          '';
        };
        locations."/metrics" = {
          # Only allow our monitoringvpn subnet
          extraConfig = ''
            allow 172.23.23.0/24;
            deny all;
          '';
          proxyPass = "http://127.0.0.1:${internalHttpPort}";
        };
        locations."/" = {
          # Return a 404 error for any paths not specified above.
          extraConfig = ''
            return 404;
          '';
        };
      };
    };

  };
}
